from ..base.base_component import ProgramHarness
from cmselemental import models

__all__ = ["GenericComponent"]


class GenericComponent(ProgramHarness):
    @classmethod
    def input(cls):
        return models.ProtoModel

    @classmethod
    def output(cls):
        return models.ProtoModel

    @staticmethod
    def found(raise_error: bool = False) -> bool:
        """
        Checks if the program can be found.
        Parameters
        ----------
        raise_error : bool, optional
            If True, raises an error if the program cannot be found.
        Returns
        -------
        bool
            Returns True if the program was found, False otherwise.
        """
        raise NotImplementedError

    def get_version(self) -> str:
        """Finds program, extracts version, returns normalized version string.
        Returns
        -------
        str
            Return a valid, safe python version string.
        """
        raise NotImplementedError



"""
Unit and regression test for the mmic package.
"""

# Import package, test suite, and other packages as needed
import mmic
import pytest
import sys


def test_mmic_imported():
    """Sample test, will always pass so long as import statement worked"""
    assert "mmic" in sys.modules


def test_mmic_generic():
    class Foo(mmic.components.blueprints.generic_component.GenericComponent):
        def execute(self):
            pass

    f = Foo(
        name="foo",
        scratch=False,
        thread_safe=False,
        thread_parallel=False,
        node_parallel=False,
        managed_memory=False,
        extras=None,
    )


def test_mmic_strategy():
    class Foo(mmic.components.blueprints.strategy_component.StrategyComponent):
        @classmethod
        def input(cls):
            return {}

        @classmethod
        def output(cls):
            return {}

        @classmethod
        def get_version(cls):
            return ""

        @classmethod
        def tactic_comps(cls):
            return set()

    f = Foo(
        name="foo",
        scratch=False,
        thread_safe=False,
        thread_parallel=False,
        node_parallel=False,
        managed_memory=False,
        extras=None,
    )


def test_mmic_tactic():
    class Foo(mmic.components.blueprints.tactic_component.TacticComponent):
        @classmethod
        def input(cls):
            return {}

        @classmethod
        def output(cls):
            return {}

        @classmethod
        def get_version(cls):
            return ""

        @classmethod
        def strategy_comp(cls):
            return ""

        def execute(self):
            pass

    f = Foo(
        name="foo",
        scratch=False,
        thread_safe=False,
        thread_parallel=False,
        node_parallel=False,
        managed_memory=False,
        extras=None,
    )



"""
TaskConfig Model
"""
from typing import Optional
import pydantic


class TaskConfig(pydantic.BaseModel):
    """Description of the configuration used to launch a task."""

    # Specifications
    ncores: int = pydantic.Field(None, description="Number cores per task on each node")
    nnodes: int  # Number of nodes per task
    memory: float  # Amount of memory in GiB per node
    scratch_directory: Optional[str]  # What location to use as scratch
    retries: int  # Number of retries on random failures
    mpiexec_command: Optional[
        str
    ]  # Command used to launch MPI tasks, see NodeDescriptor
    use_mpiexec: bool = False  # Whether it is necessary to use MPI to run an executable
    cores_per_rank: int = pydantic.Field(1, description="Number of cores per MPI rank")

    class Config:
        extra = "forbid"



from ..base.base_component import ProgramHarness
from typing import Dict, List, Set, Optional, Any, Tuple
import importlib
import abc

__all__ = ["StrategyComponent"]


class StrategyComponent(ProgramHarness):
    def execute(
        self,
        inputs: Dict[str, Any],
        extra_outfiles: Optional[List[str]] = None,
        extra_commands: Optional[List[str]] = None,
        scratch_name: Optional[str] = None,
        timeout: Optional[int] = None,
    ) -> Tuple[bool, Dict[str, Any]]:

        if isinstance(inputs, dict):
            inputs = self.input()(**inputs)

        if hasattr(inputs, "component"):
            if inputs.component:
                if not importlib.util.find_spec(inputs.component):
                    raise ModuleNotFoundError(f"{inputs.component} cannot be imported")
                else:
                    comp_mod = importlib.import_module(inputs.component)
                    return True, comp_mod._mainComponent.compute(inputs)

        if not len(self.installed_comps()):
            raise ModuleNotFoundError(
                "No supported component is installed. Solve by installing any of the following components:\n"
                + f"{self.tactic_comps}"
            )

        comp_mod = importlib.import_module(comp)
        return True, comp_mod._mainComponent.compute(inputs)

    @staticmethod
    def found(raise_error: bool = False) -> bool:
        """
        Checks if the program can be found.
        Parameters
        ----------
        raise_error : bool, optional
            If True, raises an error if the program cannot be found.
        Returns
        -------
        bool
            Returns True if the program was found, False otherwise.
        """
        raise NotImplementedError

    @classmethod
    def installed_comps(cls) -> List[str]:
        """Returns module spec if it exists.
        Returns
        -------
        List[str]
            Component names that are installed.
        """
        if cls.tactic_comps() is None:
            raise NotImplementedError(
                f"No components are available. Solve by installing any of the supported components: {cls.tactic_comps()}."
            )
        return [spec for spec in cls.tactic_comps() if importlib.util.find_spec(spec)]

    @classmethod
    @abc.abstractmethod
    def get_version(cls) -> str:
        """Finds program, extracts version, returns normalized version string.
        Returns
        -------
        str
            Return a valid, safe python version string.
        """
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def tactic_comps(cls) -> Set[str]:
        """Returns the supported tactic components e.g. set(['mmic_mda',...]).
        Returns
        -------
        Set[str]
        """
        raise NotImplementedError



import abc
from typing import Any, Dict, List, Optional, Tuple
from .config import TaskConfig
from cmselemental import models


class ProgramHarness(models.ProtoModel, metaclass=abc.ABCMeta):

    _defaults: Dict[str, Any] = {}
    name: str
    scratch: bool
    thread_safe: bool
    thread_parallel: bool
    node_parallel: bool
    managed_memory: bool
    extras: Optional[Dict[str, Any]]

    class Config:
        allow_mutation = False
        extra = "forbid"

    def __init__(self, **kwargs):
        super().__init__(**{**self._defaults, **kwargs})

    @classmethod
    @abc.abstractmethod
    def input(cls):
        ...

    @classmethod
    @abc.abstractmethod
    def output(cls):
        ...

    @classmethod
    def compute(
        cls,
        input_data: models.ProtoModel,
        config: Optional[TaskConfig] = None,
        scratch: Optional[bool] = False,
        thread_safe: Optional[bool] = False,
        thread_parallel: Optional[bool] = False,
        node_parallel: Optional[bool] = False,
        managed_memory: Optional[bool] = False,
        extras: Optional[Dict[str, Any]] = None,
    ) -> models.ProtoModel:

        # Validate the input model
        if isinstance(input_data, cls.input()):
            cls.input()(**input_data.dict())
        elif isinstance(input_data, dict):
            cls.input()(**input_data)
        else:
            raise TypeError(
                f"{type(input_data)} is not a valid input type for the {cls.__name__} component."
            )

        # Perform the execution
        program = cls(
            name=cls.__name__,
            scratch=scratch,
            thread_safe=thread_safe,
            thread_parallel=thread_parallel,
            node_parallel=node_parallel,
            managed_memory=managed_memory,
            extras=extras,
        )

        _, exec_output = program.execute(input_data)

        # Validate the output model
        if isinstance(exec_output, cls.output()):
            cls.output()(**exec_output.dict())
        elif isinstance(exec_output, dict):
            cls.output()(**exec_output)
        else:
            raise TypeError(
                f"{type(exec_output)} is not a valid output type for the {cls.__name__} component."
            )

        return exec_output

    @staticmethod
    @abc.abstractmethod
    def found(raise_error: bool = False) -> bool:
        """
        Checks if the program can be found.
        Parameters
        ----------
        raise_error : bool, optional
            If True, raises an error if the program cannot be found.
        Returns
        -------
        bool
            Returns True if the program was found, False otherwise.
        """

    ## Utility
    @abc.abstractmethod
    def get_version(self) -> str:
        """Finds program, extracts version, returns normalized version string.
        Returns
        -------
        str
            Return a valid, safe python version string.
        """
        pass

    ## Computers
    def build_input(
        self,
        input_model: models.ProtoModel,
        config: TaskConfig = None,
        template: Optional[str] = None,
    ) -> Dict[str, Any]:
        raise NotImplementedError(
            f"build_input is not implemented for {self.__class__}."
        )

    @abc.abstractmethod
    def execute(
        self,
        inputs: Dict[str, Any],
        extra_outfiles: Optional[List[str]] = None,
        extra_commands: Optional[List[str]] = None,
        scratch_name: Optional[str] = None,
        timeout: Optional[int] = None,
    ) -> Tuple[bool, Dict[str, Any]]:
        raise NotImplementedError(f"execute is not implemented for {self.__class__}.")

    def parse_output(
        self, outfiles: Dict[str, str], input_model: models.ProtoModel
    ) -> models.ProtoModel:
        raise NotImplementedError(
            f"parse_output is not implemented for {self.__class__}."
        )



"""
mmic
A short description of the project.
"""

# Add imports here
from . import components

# Handle versioneer
from ._version import get_versions

versions = get_versions()
__version__ = versions["version"]
__git_revision__ = versions["full-revisionid"]
del get_versions, versions



from ..base.base_component import ProgramHarness
import abc

__all__ = ["TacticComponent"]


class TacticComponent(ProgramHarness):
    @classmethod
    @abc.abstractmethod
    def get_version(cls) -> str:
        """Finds program, extracts version, returns normalized version string.
        Returns
        -------
        str
            Return a valid, safe python version string.
        """
        raise NotImplementedError

    @staticmethod
    def found(raise_error: bool = False) -> bool:
        """
        Checks if the program can be found.
        Parameters
        ----------
        raise_error : bool, optional
            If True, raises an error if the program cannot be found.
        Returns
        -------
        bool
            Returns True if the program was found, False otherwise.
        """
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def strategy_comp(cls) -> str:
        """Returns the strategy component this (tactic) component belongs to.
        Returns
        -------
        Set[str]
        """
        raise NotImplementedError



"""
mmic.py
A short description of the project.

Handles the primary functions
"""


def canvas(with_attribution=True):
    """
    Placeholder function to show example docstring (NumPy format)

    Replace this function and doc string for your own project

    Parameters
    ----------
    with_attribution : bool, Optional, default: True
        Set whether or not to display who the quote is from

    Returns
    -------
    quote : str
        Compiled string including quote and optional attribution
    """

    quote = "The code is but a canvas to our imagination."
    if with_attribution:
        quote += "\n\t- Adapted from Henry David Thoreau"
    return quote


if __name__ == "__main__":
    # Do something if this file is invoked on its own
    print(canvas())



"""
mmic
A short description of the project.
"""
import sys
from setuptools import setup, find_packages
import versioneer

short_description = __doc__.split("\n")

# from https://github.com/pytest-dev/pytest-runner#conditional-requirement
needs_pytest = {"pytest", "test", "ptr"}.intersection(sys.argv)
pytest_runner = ["pytest-runner"] if needs_pytest else []

try:
    with open("README.md", "r") as handle:
        long_description = handle.read()
except Exception:
    long_description = "\n".join(short_description[2:])


setup(
    # Self-descriptive entries which should always be present
    name="mmic",
    author="MolSSI",
    author_email="support@molssi.org",
    description=short_description[0],
    url="https://github.com/MolSSI/mmic",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="BSD-3-Clause",
    packages=find_packages(),
    # Optional include package data to ship with your package
    # Customize MANIFEST.in if the general case does not suit your needs
    # Comment out this line to prevent the files from being packaged with your software
    include_package_data=True,
    # Allows `setup.py test` to work correctly with pytest
    setup_requires=[] + pytest_runner,
    install_requires=['pydantic','cmselemental'],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
    ],
    # Additional entries you may want simply uncomment the lines you want and fill in the data
    # url='http://www.my_package.com',  # Website
    # install_requires=[],              # Required packages, pulls from pip if needed; do not use for Conda deployment
    # platforms=['Linux',
    #            'Mac OS-X',
    #            'Unix',
    #            'Windows'],            # Valid platforms your code works on, adjust to your flavor
    # python_requires=">=3.5",          # Python version restrictions
    # Manual control if final package is compressible or not, set False to prevent the .egg from being made
    # zip_safe=False,
)
